#!/usr/bin/env python3

import os, sys
import numpy as np
import difflib
from scipy.spatial import cKDTree
import copy
import gen, g_var, at_mod, read_in

def build_multi_residue_atomistic_system(cg_residues, sys_type):   
#### initisation of counters
    chain_count=0
    coord_atomistic={}
    g_var.seq_cg={sys_type:{}}
    g_var.ter_res={sys_type:{}}
    gen.mkdir_directory(g_var.working_dir+sys_type)  ### make and change to protein directory
#### for each residue in protein
    residue_type={}
    residue_type_mass={}
    new_chain = True
    for cg_residue_id, residue_number in enumerate(cg_residues[sys_type]):

        if np.round((cg_residue_id/len(cg_residues[sys_type]))*100,2).is_integer():
            print('Converting de_novo '+sys_type+': ',np.round((cg_residue_id/len(cg_residues[sys_type]))*100,2),'%', end='\r')
        resname = cg_residues[sys_type][residue_number][next(iter(cg_residues[sys_type][residue_number]))]['residue_name']
        if new_chain: 
            if chain_count not in coord_atomistic:
                if sys_type == 'PROTEIN':
                    g_var.backbone_coords[chain_count]=[]
                coord_atomistic[chain_count]={}
                g_var.seq_cg[sys_type][chain_count]=[]
                g_var.ter_res[sys_type][chain_count]=[resname, False]
            new_chain = False
        coord_atomistic[chain_count][residue_number]={}
        frag_location=gen.fragment_location(resname) ### get fragment location from database
        residue_type[resname], residue_type_mass[resname] = at_mod.get_atomistic(frag_location)
        g_var.seq_cg[sys_type] = add_to_sequence(g_var.seq_cg[sys_type], resname, chain_count)
        new_chain = False
        for group in residue_type[resname]:
            for key in list(residue_type[resname][group].keys()):
                if key not in cg_residues[sys_type][residue_number]:
                    del residue_type[resname][group][key]
            if len(residue_type[resname][group]) > 0:
                center, at_frag_centers, cg_frag_centers, group_fit = at_mod.rigid_fit(residue_type[resname][group], residue_type_mass[resname], residue_number, cg_residues[sys_type][residue_number])
                at_connect, cg_connect = at_mod.connectivity(cg_residues[sys_type][residue_number], at_frag_centers, cg_frag_centers, group_fit, group)
                for group_bead in group_fit:
                    if group_bead in g_var.res_top[resname]['CONNECT']:
                        at_connect, cg_connect, new_chain = at_mod.BB_connectivity(at_connect,cg_connect, cg_residues[sys_type], group_fit[group_bead], residue_number, group_bead, resname)
                        if sys_type == 'PROTEIN':
                            g_var.backbone_coords[chain_count].append(np.append(cg_residues[sys_type][residue_number][group_bead]['coord'], 1)) 
                xyz_rot_apply = at_mod.get_rotation(cg_connect, at_connect, center, resname, group, residue_number)
                coord_atomistic[chain_count] = at_mod.apply_rotations(coord_atomistic[chain_count],residue_number, group_fit, center, xyz_rot_apply)
        if new_chain:
            g_var.ter_res[sys_type][chain_count][1] = resname
            chain_count+=1

    print('Completed initial conversion of '+sys_type+'\n')        
    g_var.system[sys_type]=chain_count
    if sys_type == 'PROTEIN':
        for chain in range(chain_count):
            g_var.skip_disul[chain]=False
    return coord_atomistic

################# Fixes disulphide bond, martini cysteine bone is too far apart to be picked up by pdb2gmx. 
#### 

def ask_if_disulphide(chain, res_1, res_2):
    while True:
        try:
            answer = str(input('\nAre these residues connected by a disulphide bond in chain '+str(chain)+' (Y/N): '+str(res_1)+'--'+str(res_2)+': '))
            if answer.lower() in ['yes','y']:
                return True
            elif answer.lower() in ['no','n']:
                return False
            else:
                print("Oops!  That was a invalid choice")
        except KeyboardInterrupt:
            sys.exit('\nInterrupted')
        except BaseException:
            print("Oops!  That was a invalid choice")

def find_disulphide_bonds_user_sup():
    for chain in g_var.atomistic_protein_input_aligned:
        g_var.user_cys_bond[chain]=[]
        cysteines = []
        cys_resid = []
        for part in g_var.atomistic_protein_input_aligned[chain]:
            for resid in g_var.atomistic_protein_input_aligned[chain][part]:
                atom = next(iter(g_var.atomistic_protein_input_aligned[chain][part][resid]))
                if g_var.atomistic_protein_input_aligned[chain][part][resid][atom]['res_type'] == 'CYS':
                    for atom in g_var.atomistic_protein_input_aligned[chain][part][resid]:
                        if 'S' in g_var.atomistic_protein_input_aligned[chain][part][resid][atom]['atom'].upper() :
                            cys_resid.append(resid)
                            cysteines.append(g_var.atomistic_protein_input_aligned[chain][part][resid][atom]['coord'])
        if len(cysteines)>=2:
            tree = cKDTree(cysteines)
            done_query=[]
            for cys_index, cys in enumerate(cysteines):
                query = tree.query_ball_point(cys, r=2.3)
                if len(query) == 2 and query not in done_query:
                    g_var.user_cys_bond[chain].append([cys_resid[query[0]],cys_resid[query[1]]])
                    done_query.append(query)

def find_disulphide_bonds_de_novo():
    for chain in g_var.coord_atomistic:
        if not g_var.skip_disul[chain]:
            offset = next(iter(g_var.coord_atomistic[chain]))
            if chain not in g_var.user_cys_bond:
                g_var.user_cys_bond[chain]=[]
            cysteines = []
            cys_resid = []
            for resid in g_var.coord_atomistic[chain]:
                resid_corrected=np.array(resid)-offset
                run = True
                if len(g_var.user_cys_bond) > 0:
                    for bonds in g_var.user_cys_bond[chain]:
                        if resid_corrected in bonds:
                            run=False
                if run:
                    atom = next(iter(g_var.coord_atomistic[chain][resid]))
                    if g_var.coord_atomistic[chain][resid][atom]['res_type'] == 'CYS':
                        for atom in g_var.coord_atomistic[chain][resid]:
                            if 'S' in g_var.coord_atomistic[chain][resid][atom]['atom'].upper():
                                cys_resid.append(resid_corrected)
                                cysteines.append(g_var.coord_atomistic[chain][resid][atom]['coord'])
            if len(cys_resid)>=2:
                tree = cKDTree(cysteines)
                done_query=[]
                for cys_index, cys in enumerate(cysteines):
                    query = tree.query_ball_point(cys, r=g_var.args.cys)
                    if len(query) == 2 and query not in done_query:
                        if cys_resid[query[1]] - cys_resid[query[0]] > 4:
                            disul = True if g_var.args.silent else ask_if_disulphide(chain, cys_resid[query[0]],cys_resid[query[1]])
                            if disul:
                                g_var.user_cys_bond[chain].append([cys_resid[query[0]],cys_resid[query[1]]])
                        done_query.append(query)

def correct_disulphide_bonds(coordinates_atomistic):
    for chain in coordinates_atomistic:
        offset = next(iter(coordinates_atomistic[chain]))
        for bonds in g_var.user_cys_bond[chain]:
            bonds_corrected=np.array(bonds)+offset
            at_num=[]
            at_coord=[]
            for resid in bonds_corrected: 
                for atom in coordinates_atomistic[chain][resid]:
                    if 'S' in coordinates_atomistic[chain][resid][atom]['atom'].upper():
                        at_num.append(atom)
                        at_coord.append(coordinates_atomistic[chain][resid][atom]['coord'])
            new_c1, new_c2 = shrink_coordinates(at_coord[0],at_coord[1])
            coordinates_atomistic[chain][bonds_corrected[0]][at_num[0]]['coord']=new_c1
            coordinates_atomistic[chain][bonds_corrected[1]][at_num[1]]['coord']=new_c2
    return coordinates_atomistic

def shrink_coordinates(c1,c2):
    vector = c1-c2
    scale=0.05
    while True:
        new_c1 = c1 - (vector*scale)
        new_c2 = c2 + (vector*scale)
        dist = gen.calculate_distance(new_c1,new_c2)
        if dist > 2.1:
            scale+=0.0025
        elif dist < 1.9:
            scale-=0.001
        else:
            return new_c1, new_c2

#### disulphide checker end

### writes out final de_novo atomistic coordinates

def finalise_novo_atomistic(coord_atomistic, sys_type):
    final_at_residues={}
    final_at = {}
    for chain in coord_atomistic: 
        at_number=0    
        final_at_residues[chain]={}  
        final_at[chain]={}
        coords=[]
        skip = os.path.exists(g_var.working_dir+sys_type+'/'+sys_type+'_de_novo_'+str(chain)+'.pdb')
        if not skip:
            pdb_output = gen.create_pdb(g_var.working_dir+sys_type+'/'+sys_type+'_de_novo_'+str(chain)+'.pdb')
        for res_index, residue_id in enumerate(coord_atomistic[chain]):
            if coord_atomistic[chain][residue_id][next(iter(coord_atomistic[chain][residue_id]))]['res_type'] in g_var.mod_residues+g_var.o_residues:
                coord_atomistic[chain][residue_id] = at_mod.check_hydrogens(coord_atomistic[chain][residue_id])
            if sys_type == 'PROTEIN':
                if res_index <= len(coord_atomistic[chain])-3:
                    coord_atomistic[chain][residue_id], cross_vector = fix_carbonyl_chiral(residue_id, g_var.cg_residues['PROTEIN'], coord_atomistic[chain][residue_id], False)
                elif res_index < len(coord_atomistic[chain]) and 'cross_vector' in locals():
                    coord_atomistic[chain][residue_id], cross_vector = fix_carbonyl_chiral(residue_id, g_var.cg_residues['PROTEIN'], coord_atomistic[chain][residue_id], cross_vector)
            order = np.sort(np.array(list(coord_atomistic[chain][residue_id].keys())))
            for at_val, atom in enumerate(order):
                coord_atomistic[chain][residue_id][atom]['resid'] = res_index
                coords.append(coord_atomistic[chain][residue_id][atom]['coord'])
                final_at[chain][at_number]=coord_atomistic[chain][residue_id][atom]
                at_number+=1               
            final_at_residues[chain][res_index]=coord_atomistic[chain][residue_id]
        coords = at_mod.check_atom_overlap(coords)
        for atom in final_at[chain]:
            final_at[chain][atom]['coord']=coords[atom]
            if not skip:
                x, y, z = gen.trunc_coord(final_at[chain][atom]['coord'])
                pdb_output.write(g_var.pdbline%((atom,final_at[chain][atom]['atom'],final_at[chain][atom]['res_type'],' ',\
                                final_at[chain][atom]['resid'],x,y,z,1,0))+'\n')
    if 'pdb_output' in locals():
        pdb_output.close()
    return final_at_residues

########################################### fix carbonyl section 

def get_crossvector(cg, residue_id):
    ca=[]
    res_off = 0
    for x in range(2):
        for bead in cg[residue_id+res_off]:
            resname = cg[residue_id+res_off][next(iter(cg[residue_id+res_off]))]['residue_name']
            if bead in g_var.res_top[resname]['CONNECT']:
                if x == 0 :
                    ca.append(cg[residue_id+res_off][bead]['coord'])
                for con_val, con_dir in enumerate(g_var.res_top[resname]['CONNECT'][bead]['dir']):
                    if con_dir > 0:
                        ca.append(cg[residue_id+con_dir+res_off][g_var.res_top[resname]['CONNECT'][bead]['Con_Bd'][con_val]]['coord'])
                        res_off = con_dir      
    return ca

def get_chiral_carbonyl(at):
    carbonyl = {}
    for atom in at:
        if at[atom]['atom'] in ['N','CA', 'C', 'O']:
            carbonyl[at[atom]['atom']] = atom
    return carbonyl

def correct_carbonyl_alignment(at, cross_vector, carbonyl):
    rotation = at_mod.align_to_vector(at_mod.noramlised_vector(at[carbonyl['O']]['coord'],at[carbonyl['C']]['coord']), cross_vector)
    at[carbonyl['O']]['coord'] = (at[carbonyl['O']]['coord']-at[carbonyl['C']]['coord']).dot(rotation)+at[carbonyl['C']]['coord']
    at[carbonyl['C']]['coord'] = at[carbonyl['C']]['coord'] + cross_vector*0.2
    at[carbonyl['N']]['coord'] = at[carbonyl['N']]['coord'] - cross_vector*0.5
    return at

def fix_carbonyl_chiral(residue_id, cg, at, cross_vector):
    carbonyl = get_chiral_carbonyl(at)
    chiral = at_mod.get_chiral_non_carbonyl(at)
    if not np.any(cross_vector):
        ca = get_crossvector(cg, residue_id)
        cross_vector = at_mod.find_cross_vector( ca )
    at = correct_carbonyl_alignment(at, cross_vector, carbonyl)
#    if len(g_var.res_top[at[1]['res_type']]['CHIRAL']) >= 2:
#        at = at_mod.correct_chiral_atoms(at, chiral)
    return at, cross_vector


### align sequences of user AT structure and CG input

def add_to_sequence(sequence, residue, chain_count):
    if residue in g_var.o_residues:
        if residue in g_var.other:
            sequence[chain_count]+=g_var.other[residue]
        else:
            sequence[chain_count]+=residue
    elif residue in g_var.aas:
        sequence[chain_count]+=g_var.aas[residue]
    else:
        sequence[chain_count]+='X'    
    return sequence

def check_sequence():
    g_var.seq_at['PROTEIN'] = {}
    for chain in range(len(g_var.atomistic_protein_input_raw)):
        g_var.seq_at['PROTEIN'][chain]=[]
        for resid in g_var.atomistic_protein_input_raw[chain]:
            for atom in g_var.atomistic_protein_input_raw[chain][resid]:
                g_var.seq_at['PROTEIN'] = add_to_sequence(g_var.seq_at['PROTEIN'], g_var.atomistic_protein_input_raw[chain][resid][atom]['res_type'], chain)
                break

def align_chain_sequence(sys_type):
    cg_sequence = copy.deepcopy(g_var.seq_cg) 
    for chain_at in range(len(g_var.atomistic_protein_input_raw)):
        skip_sequence=False
        chain_cg=0
        s = difflib.SequenceMatcher(None, g_var.seq_at[sys_type][chain_at], cg_sequence[sys_type][chain_cg], autojunk=False)
        seq_info = s.get_matching_blocks()
        while seq_info[0][2] != len(g_var.seq_at[sys_type][chain_at]):
            if chain_cg >= len(cg_sequence[sys_type])-1:
                print('\nCannot find a match for user supplied chain: '+str(chain_at))#+'\n\nAtomistic chain:\n'+str(seq_user[chain_at]),'\n\nIn CG:\n'+str(sequence))
                skip_sequence = True
                break
            else:
                chain_cg+=1
                s = difflib.SequenceMatcher(None, g_var.seq_at[sys_type][chain_at], cg_sequence[sys_type][chain_cg], autojunk=False)
                seq_info = s.get_matching_blocks()
        if not skip_sequence:
            temp={}
            if chain_cg not in g_var.atomistic_protein_input_aligned:
                g_var.atomistic_protein_input_aligned[chain_cg]={}
            if seq_info[0][2] == len(g_var.seq_at[sys_type][chain_at]):
                for resid,  residue in enumerate(g_var.atomistic_protein_input_raw[chain_at]):
                    temp[resid + seq_info[0][1]] = g_var.atomistic_protein_input_raw[chain_at][residue]
                g_var.atomistic_protein_input_aligned[chain_cg][str(seq_info[0][1])+':'+str(seq_info[0][1]+seq_info[0][2])]=temp  
            cg_sequence[sys_type][chain_cg] = mask_sequence(cg_sequence[sys_type][chain_cg], seq_info[0][1], seq_info[0][1]+seq_info[0][2])
            g_var.cg_chain_group[chain_at]=chain_cg

    check_chain_alignment_coverage(sys_type)

    if len(g_var.atomistic_protein_input_raw) < len(g_var.seq_cg[sys_type]):
        print('### WARNING you have supplied fewer chains than exist in the CG system ###\n')
    if len(g_var.atomistic_protein_input_aligned) > 0:
        g_var.user_at_input = True
    else: 
        g_var.user_at_input = False
    sort_chains(sys_type)
    if g_var.args.v >= 2:
        print(gen.print_sequnce_info('PROTEIN'))

def sort_chains(sys_type):
    if g_var.group_chains is None:
        g_var.group_chains = {}
        for i in range(len(g_var.seq_at[sys_type])):
            g_var.group_chains[i] = i
    elif g_var.group_chains == 'chain':
        g_var.group_chains = g_var.cg_chain_group
    elif g_var.group_chains == 'all':
        g_var.group_chains = {}
        for i in range(len(g_var.cg_chain_group)):
            g_var.group_chains[i] = 0
    else:
        sys.exit('Failed to parse chain sorting')


def check_chain_alignment_coverage(sys_type):
    for key, chain in g_var.atomistic_protein_input_aligned.items():
        total = 0
        for fragment in chain:
            resid_range = fragment.split(':')
            total+=(int(resid_range[1])-int(resid_range[0]))
        if total == len(g_var.seq_cg[sys_type][key]):
            g_var.skip_disul[key]=True
        else:
            g_var.skip_disul[key]=False

def mask_sequence(sequence, st, end):
    for index, residue in enumerate(sequence):
        if st <= index < end:
            sequence[index]='-'
    return sequence

#####   aligning user protein chains

def align_user_chains(final_coordinates_atomistic):
    atomistic_protein_centered, cg_com = center_atomistic() ## centers each chain by center of mass
    at_com_group, cg_com_group = rotate_protein_monomers(atomistic_protein_centered, final_coordinates_atomistic, cg_com) 
    atomistic_protein_rotated = apply_rotations_to_chains(final_coordinates_atomistic, atomistic_protein_centered, 
                                                                at_com_group,cg_com_group,cg_com) ## apply rotation matrix to atoms and build in missing residues
    final_user_supplied_coord = correct_disulphide_bonds(atomistic_protein_rotated) ## fixes sulphur distances in user structure
    for chain, protein in final_user_supplied_coord.items():
        at_id=0
        coord=[]
        final_atom = []
        for resid, residue in protein.items():
            for atom in residue.values():
                coord.append(atom['coord'])
                final_atom.append({'atom_name':atom['atom'], 'residue_name':atom['res_type'], 'chain':chain, 'residue_id':resid,\
                                    'x':atom['coord'][0],'y':atom['coord'][1],'z':atom['coord'][2]})
                at_id+=1
        corrected_coords, index_conversion = at_mod.index_conversion_generate(final_atom, coord)
        at_mod.write_pdb(final_atom, corrected_coords, index_conversion, g_var.working_dir+'PROTEIN/PROTEIN_aligned_'+str(chain)+'.pdb')

def center_atomistic():
    cg_com={}
#### for each protein chain center on cg representation 
    for chain in g_var.atomistic_protein_input_aligned:
        cg_com[chain]=[]
        for part_val, part in enumerate(g_var.atomistic_protein_input_aligned[chain]):
            sls, sle= int(part.split(':')[0]),int(part.split(':')[1])
            if 'protein_mass' not in locals():
                protein_mass={}
            if chain in g_var.group_chains:
                if g_var.group_chains[chain] not in protein_mass:
                    protein_mass[g_var.group_chains[chain]]=fetch_backbone_mass(g_var.atomistic_protein_input_aligned[chain][part], [])
                else:
                    protein_mass[g_var.group_chains[chain]]=fetch_backbone_mass(g_var.atomistic_protein_input_aligned[chain][part], protein_mass[g_var.group_chains[chain]])
                if 'cg_backbone_masses' not in locals():
                    cg_backbone_masses = {}
                if g_var.group_chains[chain] not in cg_backbone_masses:
                    cg_backbone_masses[g_var.group_chains[chain]]= g_var.backbone_coords[chain][sls:sle]
                else:
                    cg_backbone_masses[g_var.group_chains[chain]] += g_var.backbone_coords[chain][sls:sle]                  
    g_var.atomistic_protein_input_aligned, cg_com = center_at_protein_chain_groups(g_var.atomistic_protein_input_aligned, cg_com, protein_mass, cg_backbone_masses)
    return g_var.atomistic_protein_input_aligned, cg_com

def center_at_protein_chain_groups(atomistic_protein_input, cg_com, protein_mass, cg_backbone_masses):
    for chain in atomistic_protein_input:
        for part_val, part in enumerate(atomistic_protein_input[chain]):
            sls, sle= int(part.split(':')[0]),int(part.split(':')[1])
            if chain in g_var.group_chains:
                atomistic_protein_mass = at_mod.COM(protein_mass[g_var.group_chains[chain]], 'AT protein chain: '+str(chain))
                cg_com[chain].append(at_mod.COM(cg_backbone_masses[g_var.group_chains[chain]], 'CG protein chain: '+str(chain)))
            else:
                protein_mass=fetch_backbone_mass(atomistic_protein_input[chain][part], [])
                atomistic_protein_mass = at_mod.COM(protein_mass, 'protein at: '+str(chain)+' '+part)#### add center of mass of CG_proteins
                cg_com[chain].append(at_mod.COM(g_var.backbone_coords[chain][sls:sle], 'protein cg: '+str(chain)+' '+part))
            atomistic_protein_input[chain][part] = update_part_coordinate(atomistic_protein_input[chain][part], atomistic_protein_mass, cg_com[chain][part_val])
    return atomistic_protein_input, cg_com

def update_part_coordinate(part, at_com, cg_com):
    #### each atoms coord is updated so the AT COM is the same as the CG
    for residue in part:
        for atom in part[residue]:  
            part[residue][atom]['coord']=part[residue][atom]['coord']-(at_com-cg_com)
    return part

def fetch_backbone_mass(part, protein_mass):
    for residue in part:
    #### creates a list of all coordinates and masses [[coord, mass],[coord, mass]]
        for atom in part[residue]:
            if part[residue][atom]['atom'] in g_var.res_top[part[residue][atom]['res_type']]['ATOMS']:
                protein_mass.append([part[residue][atom]['coord'][0],part[residue][atom]['coord'][1],part[residue][atom]['coord'][2],part[residue][atom]['frag_mass']])    
    return protein_mass

def rotate_protein_monomers(atomistic_protein_centered, final_coordinates_atomistic, cg_com):
#### run through each chain in proteins
    at_com_group, cg_com_group={},{}
    for chain in range(len(final_coordinates_atomistic)):
    #### creates atomistic pdb
        if chain in atomistic_protein_centered:
            for part_val, part in enumerate(atomistic_protein_centered[chain]):
                sls, sle= int(part.split(':')[0]),int(part.split(':')[1])        
                at_centers=[]
            #### runs through every residue and atom  
                for residue in atomistic_protein_centered[chain][part]:
                #### gets center of mass of each residue (note only backbone heavy atoms have a mass)
                    at_centers_iter=[]

                    for atom in atomistic_protein_centered[chain][part][residue]:

                        if atomistic_protein_centered[chain][part][residue][atom]['atom'] in g_var.res_top[atomistic_protein_centered[chain][part][residue][atom]['res_type']]['ATOMS']:
                            at_centers_iter.append(np.append(atomistic_protein_centered[chain][part][residue][atom]['coord'],atomistic_protein_centered[chain][part][residue][atom]['frag_mass']))
                    try:
                        at_centers.append(np.average(np.array(at_centers_iter)[:,:3], axis=0, weights=np.array(at_centers_iter)[:,3]))
                    except BaseException:
                        for atom in atomistic_protein_centered[chain][part][residue]:
                            print(atomistic_protein_centered[chain][part][residue][atom])
                        sys.exit()
                if len(at_centers) == len(np.array(g_var.backbone_coords[chain])[sls:sle,:3]):
                    at_com_group, cg_com_group = return_grouped_rotations(chain, part_val, at_centers, cg_com, at_com_group, cg_com_group, sls, sle)
                else:
                    sys.exit('In chain '+str(chain)+' the atomistic input does not match the CG. \n\
                            number of CG residues '+str(len(g_var.backbone_coords[chain]))+'\nnumber of AT residues '+str(len(at_centers)))
    return at_com_group, cg_com_group

def apply_rotations_to_chains(final_coordinates_atomistic, atomistic_protein_centered, at_com_group,cg_com_group,cg_com):
    final_user_supplied_coord={}
    for chain in range(len(final_coordinates_atomistic)):
        rotations=[]
        if chain in atomistic_protein_centered:
            rotate_chain=at_mod.kabsch_rotate(at_com_group[g_var.group_chains[chain]]-cg_com[g_var.group_chains[chain]][0], cg_com_group[g_var.group_chains[chain]]-cg_com[g_var.group_chains[chain]][0])
            for part_val, part in enumerate(atomistic_protein_centered[chain]):
                if chain in g_var.group_chains:
                    rotations.append(rotate_chain)
                else:
                    rotations = at_com_group[chain]
            final_user_supplied_coord[chain] = hybridise_protein_inputs(final_coordinates_atomistic[chain], atomistic_protein_centered[chain], cg_com[chain], rotations, chain)
        else:
            final_user_supplied_coord[chain] = hybridise_protein_inputs(final_coordinates_atomistic[chain], [], [], [], chain)
    return final_user_supplied_coord

def return_indivdual_rotations(chain, part_val, at_centers, cg_com, at_com_group, cg_com_group, sls, sle):
    if chain not in at_com_group:
        at_com_group[chain]=[]
    if len(at_centers) == len(np.array(g_var.backbone_coords[chain])[sls:sle,:3]):
        at_com_group[chain].append(at_mod.kabsch_rotate(np.array(at_centers)-cg_com[chain][part_val], 
                             np.array(g_var.backbone_coords[chain])[sls:sle,:3]-cg_com[chain][part_val]))
    else:
        sys.exit('In chain '+str(chain)+' the atomistic input does not match the CG. \n\
                    number of CG residues '+str(len(g_var.backbone_coords[chain]))+'\nnumber of AT residues '+str(len(at_centers)))    
    return at_com_group

def return_grouped_rotations(chain, part_val, at_centers, cg_com, at_com_group, cg_com_group, sls, sle):
    if chain in g_var.group_chains:
        if g_var.group_chains[chain] not in at_com_group:
            at_com_group[g_var.group_chains[chain]]=[]   
            at_com_group[g_var.group_chains[chain]]=np.array(at_centers)
            cg_com_group[g_var.group_chains[chain]]=np.array(g_var.backbone_coords[chain])[sls:sle,:3]  
        else:
            at_com_group[g_var.group_chains[chain]]=np.append(at_com_group[g_var.group_chains[chain]], np.array(at_centers), axis=0)
            cg_com_group[g_var.group_chains[chain]]=np.append(cg_com_group[g_var.group_chains[chain]], np.array(g_var.backbone_coords[chain])[sls:sle,:3], axis=0)                                                                               
    else:
        at_com_group = return_indivdual_rotations(chain, part_val, at_centers, cg_com, at_com_group, cg_com_group, sls, sle)
    return at_com_group, cg_com_group

def hybridise_protein_inputs(final_coordinates_atomistic, atomistic_protein_centered, cg_com, xyz_rot_apply, chain):
    complete_user_at = {}
    for residue in final_coordinates_atomistic:
        exists=False
        resname = final_coordinates_atomistic[residue][next(iter(final_coordinates_atomistic[residue]))]['res_type']
        if resname in g_var.mod_residues:
            complete_user_at[residue]=final_coordinates_atomistic[residue]
        elif resname not in g_var.mod_residues:
            for part_val, part in enumerate(atomistic_protein_centered):
                if residue in atomistic_protein_centered[part]:
                    exists=True
                    for atom in atomistic_protein_centered[part][residue]:   
                        if atomistic_protein_centered[part][residue][atom]['res_type'] != resname:
                            print('de_novo' , resname,'at_user', atomistic_protein_centered[part][residue][atom]['res_type'])
                            sys.exit('de novo and at user supplied don\'t match')
                        atomistic_protein_centered[part][residue][atom]['coord'] = at_mod.rotate_atom(atomistic_protein_centered[part][residue][atom]['coord'], cg_com[part_val], xyz_rot_apply[part_val])
                    complete_user_at[residue]=atomistic_protein_centered[part][residue]
        if not exists:
            complete_user_at[residue]=final_coordinates_atomistic[residue]
    return complete_user_at

################################################################## Merge chains ####################    

def correct_amide_h(lines, coords):
    for at_val, atom in enumerate(lines):
        if atom['atom_name'] == 'C': 
            C = np.array(coords[at_val])
        if atom['atom_name'] == 'O':   
            O = np.array(coords[at_val]) 

        resname = gen.check_alternate_resname(atom['residue_name'])
        if atom['atom_name'] == g_var.res_top[resname]['amide_h'] and atom['residue_id'] != 0:  
            HN = np.array(coords[at_val]) 
            HN_index = at_val
        if atom['atom_name'] == 'N' and atom['residue_id'] != 0:
            N = np.array(coords[at_val])
        if 'C' in locals() and 'N' in locals() and 'O' in locals() and 'HN' in locals():
            O_C = O-C
            coords[at_val] = N - O_C
            del N, C, O, HN 
    return coords

def create_disres(coord, chain, file, at_start, count):
    P_R = np.array([])
    if chain in g_var.atomistic_protein_input_aligned:
        for key in g_var.atomistic_protein_input_aligned[chain].keys():
            P_R = np.append(P_R, np.arange(int(key.split(':')[0]), int(key.split(':')[1])+1))
        HN, O = get_backbone(coord)
        count, disres = find_connect(coord, P_R, HN, O, at_start, count)
        write_disres(disres)
    return len(coord)+at_start, count

def write_disres(disres):        
    header = True if not os.path.exists(g_var.working_dir+'PROTEIN/PROTEIN_disres.itp') else False
    with open(g_var.working_dir+'PROTEIN/PROTEIN_disres.itp', 'a') as disres_out:
        if header:
            disres_out.write(';backbone hydrogen bonding distance restraints\n\n')
            disres_out.write('[ intermolecular_interactions ]\n[ distance_restraints ]\n')
            disres_out.write(';   i     j type label      funct         lo        up1        up2     weight\n')
        for restraint in disres:
            disres_out.write(restraint)

def get_backbone(coord):
    HN, O = [],[]
    for atom in coord:
        if atom['residue_name'] in g_var.p_residues and atom['atom_name'] == g_var.res_top[atom['residue_name']]['amide_h']:
            HN.append([int(atom['atom_number']), atom['x'], atom['y'], atom['z']])
        if atom['residue_name'] in g_var.p_residues and atom['atom_name'] == 'O':
            O.append([int(atom['atom_number']), atom['x'], atom['y'], atom['z']])
    return np.array(HN), np.array(O)     
            
def find_connect(coord, P_R, HN, O, at_start, count):
    disres = []
    tree = cKDTree(HN[:,1:])
    for carbonyl in O:
        ndx = tree.query_ball_point(carbonyl[1:], r=3)
        if len(ndx) > 0:
            for at in ndx:
                HN_resid = coord[int(HN[at][0])-1]['residue_id']
                O_resid = coord[int(carbonyl[0])-1]['residue_id']
                if HN_resid in P_R and O_resid in P_R and HN_resid not in [O_resid-1 ,O_resid, O_resid+1]:
                    count+=1
                    xyz1 = [coord[int(carbonyl[0])-1]['x'], coord[int(carbonyl[0])-1]['y'], coord[int(carbonyl[0])-1]['z']]
                    xyz2 = [coord[int(HN[at][0])-1]['x'], coord[int(HN[at][0])-1]['y'], coord[int(HN[at][0])-1]['z']]
                    dist = (gen.calculate_distance(xyz1, xyz2)/10)-0.05
                    disres.append('{0:10}{1:10}{2:3}{3:12}{4:12}{5:^12}{6:14}{7:14}{8:5}\n'.format(str(at_start+int(HN[at][0])), str(at_start+int(carbonyl[0])), 
                                                                                                         '1', str(count),'2', '0', str(np.round(dist,4)), str(np.round(dist+0.01,4)), '1'))
    return count, disres 

########################################################### RMSD

def write_RMSD():
    print('\nCalculating backbone RMSDs\n')
    RMSD={}
    de_novo_atoms, chain_count = read_in.read_in_atomistic(g_var.final_dir+'final_cg2at_de_novo.pdb') ## reads in final pdb
    if chain_count != g_var.system['PROTEIN']:
        sys.exit('number of chains in atomistic protein input ('+str(chain_count)+') does not match CG representation ('+str(g_var.system['PROTEIN'])+')')
    RMSD_de_novo = RMSD_measure_de_novo(de_novo_atoms) ## gets rmsd of de novo

    if g_var.user_at_input and 'PROTEIN' in g_var.cg_residues: 
        if g_var.args.o in ['all', 'align']: 
            at_input_atoms, chain_count = read_in.read_in_atomistic(g_var.final_dir+'final_cg2at_aligned.pdb')
            RMSD_aligned = RMSD_measure_de_novo(at_input_atoms)   
            seg_rmsd = RMSD_measure_aligned(at_input_atoms)   

    with open(g_var.final_dir+'structure_quality.dat', 'w') as qual_out:   
        line_1   = ' chain    De novo BB RMSD ('+chr(197)+')'
        line_2   = ' -----    -------------------'
        if 'seg_rmsd' in locals():
            line_1+= '    Aligned BB RMSD ('+chr(197)+')    Seg backbone RMSD ('+chr(197)+')'
            line_2+= '    -------------------    ---------------------'
        qual_out.write(line_1+'\n'+line_2+'\n')
        print(line_1+'\n'+line_2)
        write_warning = False
        for chain in RMSD_de_novo:
            line = ' {0:^5}{1:^28}'.format(str(chain), float(RMSD_de_novo[chain]))
            if 'seg_rmsd' in locals():
                line += '{0:^18}'.format(float(RMSD_aligned[chain]))
                if chain in seg_rmsd:
                    line+= '{0:^28}'.format(', '.join(seg_rmsd[chain])) 
                    if '*' in ', '.join(seg_rmsd[chain]):
                        write_warning = True
            print(line)
            qual_out.write(line+'\n')
        if write_warning:
            print('\n * Segment alignments may have minor deviations due to either clashes or structure hybridisation.')
            qual_out.write('\n * Segment alignments may have minor deviations due to either clashes or structure hybridisation.')
        print('\n\nAll RMSDs have been saved in: \n'+g_var.final_dir+'structure_quality.dat\n')

def RMSD_measure_de_novo(structure_atoms):
    RMSD_dict = {}
    for chain in range(g_var.system['PROTEIN']):
        at_centers=[]
    #### runs through every residue and atom  
        for residue in structure_atoms[chain]:
        #### gets center of mass of each residue (note only backbone heavy atoms have a mass)
            at_centers_iter=[]
            for atom in structure_atoms[chain][residue].values():
                if atom['atom'] in g_var.res_top[atom['res_type']]['ATOMS']:
                    at_centers_iter.append(np.append(atom['coord'],atom['frag_mass']))
            if np.any(np.array(at_centers_iter)[:,3] > 0):
                at_centers.append(np.average(np.array(at_centers_iter)[:,:3], axis=0, weights=np.array(at_centers_iter)[:,3]))
            else:
                sys.exit('Missing masses for RMSD')

    #### checks that the number of residues in the chain are the same between CG and AT
        if len(at_centers) != len(g_var.backbone_coords[chain]):
            sys.exit('In chain '+str(chain)+' the atomistic input does not match the CG. \n\
    number of CG residues '+str(len(g_var.backbone_coords[chain]))+'\nnumber of AT residues '+str(len(at_centers)))

        cg_center = np.mean(np.array(g_var.backbone_coords[chain])[:,:3], axis=0)
        at_align = np.array(at_centers) - (np.mean(np.array(at_centers), axis=0) - cg_center)
        at_align = RMSD_align(at_align, np.array(g_var.backbone_coords[chain])[:,:3])
        RMSD_dict[chain]=Calculate_RMSD(np.array(at_align), np.array(g_var.backbone_coords[chain])[:,:3])  #### stores RMSD in dictionary
    return RMSD_dict

def RMSD_align(coord_set_1, coord_set_2):
    center = np.mean(coord_set_2, axis=0)
    xyz_rot_apply=at_mod.kabsch_rotate(coord_set_1-center, coord_set_2-center)
    ali= []
    for at_val, atom in enumerate(coord_set_1):
        ali.append( at_mod.rotate_atom(atom, center, xyz_rot_apply) )
    return np.array(ali)

def Calculate_RMSD(C1, C2):
    dist=np.sqrt((C1 - C2)**2)
    return np.round(np.sqrt(np.mean(dist**2)),3) #### RMSD calculation

def get_coordinates(input_coord, P_R, chain):
    coord_dict = {}
    for residue_val, residue in input_coord[chain].items():
        for seg_val, seg in enumerate(P_R):
            if seg_val not in coord_dict:
                coord_dict[seg_val]=[]
            if residue_val in seg:
                for atom in residue.values():
                    if atom['atom'] in g_var.res_top[atom['res_type']]['ATOMS'] and atom['res_type'] in g_var.p_residues:
                        coord_dict[seg_val].append(np.append(atom['coord'],atom['frag_mass']))
                break
    for segment in range(len(coord_dict)):
        coord_dict[segment] = np.array(coord_dict[segment])
    return coord_dict

def RMSD_measure_aligned(Final_structure):
    seg_rmsd={}
    total_initial, total_final = np.array([]),np.array([])
    for chain in g_var.atomistic_protein_input_aligned:
        initial_structure, chain_count = read_in.read_in_atomistic(g_var.working_dir+'PROTEIN/MIN/PROTEIN_aligned_'+str(chain)+'.pdb')
        P_R = []
        for key in g_var.atomistic_protein_input_aligned[chain].keys():
            P_R.append(np.arange(int(key.split(':')[0])-1, int(key.split(':')[1])))
        final_backbone = get_coordinates(Final_structure, P_R, chain)
        initial_backbone = get_coordinates(initial_structure, P_R, 0)
        seg_rmsd[chain]=[]
        for segment in range(len(final_backbone)):
            initial_backbone_fitted = RMSD_align(initial_backbone[segment][:,:3],final_backbone[segment][:,:3])
            total_initial = np.append(total_initial, initial_backbone_fitted)
            total_final = np.append(total_final, final_backbone[segment][:,:3])
            RMSD_val = Calculate_RMSD(np.array(initial_backbone_fitted), final_backbone[segment][:,:3])
            if RMSD_val > 0.15:
                seg_rmsd[chain].append(str(RMSD_val)+' *')
            else:
                seg_rmsd[chain].append(str(RMSD_val))
    # print(Calculate_RMSD(total_initial, total_final))
    return seg_rmsd



#!/usr/bin/env python3

import os, sys
import numpy as np
import time
import multiprocessing as mp
sys.path.append(os.path.dirname(os.path.realpath(__file__))+'/database/bin')
import gen, gro, at_mod, at_mod_p, at_mod_np, read_in, g_var, check_library


if __name__ == '__main__':
    mp.freeze_support()
    ## hardcoded varibles used by the script
    ## I've tried to make them as comprehensive as possible but they may need updating occasionally
    g_var.version = '0.2'
    
    g_var.script_update = '28-07-2025'

    g_var.other = {'DA':'A', 'DG':'G', 'DC':'C', 'DT':'T', 'DAX':'A', 'DGX':'G', 'DCX':'C', 'DTX':'T'}

# Nothing in the script below here should need changing by the user
    
    g_var.tc['i_t']=time.time()
    ### initialise script 
    gen.cg2at_header()
    gen.fetch_forcefield_water_info()
    gen.check_input_flag() #### if missing structure file print help and quit
    gen.correct_number_cpus()
    gen.find_gromacs()
    gen.read_database_directories()
    gen.forcefield_selection()
    gen.fragment_selection()
    gen.check_water_molecules()
    if g_var.args.posre != None and len(g_var.np_directories) > 0:
        check_library.add_posres_file()
    if g_var.args.compare != None and len(g_var.np_directories) > 0:
        check_library.compare_forcefield_to_database()
    if g_var.args.info:
        gen.database_information()
    if g_var.args.v >= 1:
        print(gen.fragments_in_use())
    gen.get_termini_selections()
    gen.fetch_fragment_multi() 
    gen.fetch_fragment_single()   
    gen.fetch_chain_groups()
    gen.sort_swap_group()
    print(gen.print_swap_residues())
    ###
    #### collects initial structures into INPUT folder
    gro.collect_input()
    #### saves flags used into INPUT folder
    gen.flags_used()
    g_var.tc['i_t_e']=time.time()
    #### reads in CG file and separates into residue types
    box_vec_initial = read_in.read_initial_cg_pdb()
    #### box size update 
    if g_var.args.box != None:
        print('box cutting only works for cubic boxes currently')
        g_var.box_vec, box_shift = gen.new_box_vec(box_vec_initial, g_var.args.box)
    else:
        g_var.box_vec=box_vec_initial
        box_shift=np.array([0,0,0])
    read_in.real_box_vectors(g_var.box_vec)
    #### pbc fix and residue truncation if required
    read_in.fix_pbc(box_vec_initial, g_var.box_vec, box_shift)
    #### checks if fragment database and input files match  
    at_mod.sanity_check()
    ### convert protein to atomistic representation
    g_var.tc['r_i_t']=time.time()
    if 'PROTEIN' in g_var.cg_residues:          
        g_var.coord_atomistic = at_mod_p.build_multi_residue_atomistic_system(g_var.cg_residues, 'PROTEIN') ## converts protein to atomistic
        if not g_var.user_at_input and g_var.args.v >= 1:  ## prints protein sequences 
            print(gen.print_sequnce_info('PROTEIN'))
        ## reads in user chain, runs a sequence alignment and finds existing disulphide bonds
        g_var.tc['p_d_n_t']=time.time()
        if g_var.user_at_input:
            for file_num, file_name in enumerate(g_var.args.a):
                atomistic_protein_input_raw, g_var.chain_count = read_in.read_in_atomistic(g_var.input_directory+'AT_INPUT_'+str(file_num)+'.pdb')  ## reads in user structure
                g_var.atomistic_protein_input_raw.update(atomistic_protein_input_raw)
            read_in.duplicate_chain()  ## duplicates user chcains
            at_mod_p.check_sequence()  ## gets user sequence
            at_mod_p.align_chain_sequence('PROTEIN') ## aligns chains 
            at_mod_p.find_disulphide_bonds_user_sup() ## finds user disulphide bonds
        at_mod_p.find_disulphide_bonds_de_novo() ## finds CG disulphide bonds 
        g_var.coord_atomistic = at_mod_p.correct_disulphide_bonds(g_var.coord_atomistic) ## fixes sulphur distances
        final_coordinates_atomistic_de_novo = at_mod_p.finalise_novo_atomistic(g_var.coord_atomistic, 'PROTEIN') ## fixes carbonyl oxygens, hydrogens and writes pdb 
        ## aligns user chains to the CG system
        if g_var.user_at_input:
            at_mod_p.align_user_chains(final_coordinates_atomistic_de_novo)
        #### read in minimised de novo protein chains and merges chains
        if not os.path.exists(g_var.working_dir+'PROTEIN/PROTEIN_de_novo_merged.pdb'):
            gro.run_parallel_pdb2gmx_min('PROTEIN', g_var.ter_res['PROTEIN'])### runs pdb2gmx and minimises each protein chain
            print('Merging de_novo protein chains')
            at_mod.merge_indivdual_chain_pdbs(g_var.working_dir+'PROTEIN/MIN/PROTEIN_de_novo', '.pdb', 'PROTEIN') ## merge protein chains

        #### read in aligned protein chains and merges chains
        if g_var.user_at_input and not os.path.exists(g_var.working_dir+'PROTEIN/PROTEIN_aligned_merged.pdb'):
            print('Merging aligned protein chains')
            if g_var.args.o not in ['none', 'align']:
                at_mod.merge_indivdual_chain_pdbs(g_var.working_dir+'PROTEIN/MIN/PROTEIN_aligned', '.pdb', 'PROTEIN') ## merge aligned chains
            else:
                at_mod.merge_indivdual_chain_pdbs(g_var.working_dir+'PROTEIN/PROTEIN_aligned', '_gmx_checked.pdb', 'PROTEIN')

    ### converts other linked residues  
    g_var.tc['f_p_t']=time.time()
    if 'OTHER' in g_var.cg_residues:  
        g_var.other_atomistic = at_mod_p.build_multi_residue_atomistic_system(g_var.cg_residues, 'OTHER')   
        if g_var.args.v >= 1:  ## prints protein sequences 
            print(gen.print_sequnce_info('OTHER'))   
        fin_at_NP_linked_de_novo = at_mod_p.finalise_novo_atomistic(g_var.other_atomistic, 'OTHER')
        gro.run_parallel_pdb2gmx_min('OTHER', g_var.ter_res['OTHER'])
        if not os.path.exists(g_var.working_dir+'OTHER/OTHER_de_novo_merged.pdb'):
            at_mod.merge_indivdual_chain_pdbs(g_var.working_dir+'OTHER/MIN/OTHER_de_novo', '.pdb', 'OTHER') ## merge  chains
    g_var.tc['f_o_t']=time.time()

    #### converts non protein residues into atomistic (runs on all cores)
    if len([key for value, key in enumerate(g_var.cg_residues) if key not in ['PROTEIN', 'OTHER']]) > 0:
        print('\nConverting the following residues: \n')
        # os.chdir(g_var.start_dir)
        for residue_type in g_var.cg_residues.keys():
            if residue_type not in ['PROTEIN', 'OTHER']:
                number = at_mod_np.build_atomistic_system(residue_type) 
                g_var.system.update(number)
        # with mp.Pool(g_var.args.ncpus) as pool:
        #     pool_process = pool.starmap_async(at_mod_np.build_atomistic_system, [(residue_type, g_var.working_dir, g_var.cg_residues) 
        #                                 for residue_type in [key for key in g_var.cg_residues if key not in ['PROTEIN', 'OTHER']]]).get() ## fragment fitting done in parrallel  
        # for residue_type in pool_process:
        #     g_var.system.update(residue_type) ## updates residue counts 
        #### attempts to minimise all residues at once 
        print('\nThis may take some time....(probably time for a coffee)\n')
        for residue_type in [key for key in g_var.cg_residues if key not in ['PROTEIN', 'OTHER']]:
            if not os.path.exists(g_var.working_dir+residue_type+'/'+residue_type+'_merged.pdb'):
                print('Minimising: '+residue_type) 
                error = gro.minimise_merged(residue_type, g_var.working_dir+residue_type+'/'+residue_type+'_all.pdb')
                if error == True and residue_type not in ['SOL']:
                    print('Failed to minimise as a group: '+residue_type)
                    print('please check your input file as there is likely something wrong')
 
    ### MERGES system
    g_var.tc['n_p_t']=time.time()

    print('Merging all residue types to single file. (Or possibly tea)\n')

    gro.write_merged_topol() ## make final topology in merged directory

    #### copies all itp files and topologies from wherever they are stored into the FINAL folder
    for file_name in os.listdir(g_var.merged_directory):
        if not any(f in file_name for f in ['steered_posre.itp', 'low_posre.itp','mid_posre.itp', 'high_posre.itp']):
            if file_name.endswith('.itp') or file_name.endswith('final.top') :
               gen.file_copy_and_check(g_var.merged_directory+file_name, g_var.final_dir+file_name)

    #### merges provided atomistic protein and residues types into a single pdb file into merged directory
    if not os.path.exists(g_var.merged_directory+'merged_cg2at_de_novo.pdb'):
        at_mod.merge_system_pdbs('_de_novo') ## merge all minimised residues into a complete system 

    ## minimise merged system
    if not os.path.exists(g_var.merged_directory+'MIN/merged_cg2at_de_novo_minimised.pdb'):
        gro.make_min('merged_cg2at') 
        gro.minimise_merged_pdbs( '_de_novo') ## minimise system pdb
    g_var.tc['m_t']=time.time()
    ## checks for threaded lipids, e.g. abnormal bonds lengths (not had a issue for a long time might delete) 
    if not os.path.exists(g_var.merged_directory+'checked_ringed_lipid_de_novo.pdb'):
        at_mod.check_ringed_lipids(g_var.merged_directory+'MIN/merged_cg2at_de_novo_minimised.pdb') ## check for abnormal bond lengths 
    ## runs short NVT on de_novo system with disres on if available 
    if g_var.args.o not in ['none', 'align']:
        gro.run_nvt(g_var.merged_directory+'checked_ringed_lipid_de_novo') ## run npt on system 
    else:
        print('Completed initial minimisation, please find final de_novo system: \n'+g_var.final_dir+'final_cg2at_de_novo.pdb')
        gen.file_copy_and_check(g_var.merged_directory+'checked_ringed_lipid_de_novo.pdb', g_var.final_dir+'final_cg2at_de_novo.pdb')
    g_var.tc['eq_t']=time.time()
    ## creates aligned system 

    if g_var.user_at_input and 'PROTEIN' in g_var.cg_residues and g_var.args.o in ['all', 'align']:   
        g_var.tc['a_s']=time.time()
        gro.create_aligned()
        g_var.tc['a_e']=time.time()

    ## removes temp file from script
    if not g_var.args.messy:
        gen.clean() 

    ## prints out system information
    print(gen.write_system_components())

    ## prints out RMSD of converted proteins
    if 'PROTEIN' in g_var.cg_residues:
        at_mod_p.write_RMSD()

    g_var.tc['f_t']=time.time()

    ## prints out script timings for each section
    gen.print_script_timings()






#!/usr/bin/env python3

import os, sys
import numpy as np
import subprocess 
import multiprocessing as mp
from shutil import rmtree
import time
import gen, g_var, at_mod, read_in, at_mod_p

terminal_PTMs = ['CYST', 'GLYM']


#### collects input structures and creates initial folders
def collect_input():
    if os.path.exists(g_var.args.c):
        print('You have selected the following coarse-grained input file: ', g_var.args.c)
    else:
        sys.exit('Cannot find CG input file: '+g_var.args.c)
    gen.mkdir_directory(g_var.working_dir)
    gen.mkdir_directory(g_var.final_dir)
    gen.mkdir_directory(g_var.input_directory)
    gen.mkdir_directory(g_var.merged_directory)
#### collates all input files in input directory
    if g_var.args.a != None:
        print('You have selected the following atomistic input files:') 
        for file_num, file_name in enumerate(g_var.args.a):
            if os.path.exists(file_name):
                print( file_name )
            else:
                sys.exit('cannot find atomistic input file: '+file_name)
            gen.file_copy_and_check(file_name, g_var.input_directory+gen.path_leaf(file_name)[1])
            os.chdir(g_var.input_directory)
            gromacs([g_var.args.gmx+' editconf -f '+gen.path_leaf(file_name)[1]+' -resnr 0 -o '+g_var.input_directory+'AT_INPUT_'+str(file_num)+'.pdb', g_var.input_directory+'AT_INPUT_'+str(file_num)+'.pdb'])
            if not os.path.exists(g_var.input_directory+'AT_INPUT_'+str(file_num)+'.pdb'):
                sys.exit('\nFailed to process atomistic input file.\nCheck gromacs outputs in '+g_var.input_directory)
            else:
                g_var.user_at_input = True
            os.chdir(g_var.start_dir)

    gen.file_copy_and_check(g_var.args.c, g_var.input_directory+gen.path_leaf(g_var.args.c)[1])
    os.chdir(g_var.input_directory)
    if not hasattr(g_var, 'gmx_version'):
        gromacs([g_var.args.gmx+' -version', 'version.txt'])
    gromacs([g_var.args.gmx+' editconf -f '+gen.path_leaf(g_var.args.c)[1]+' -resnr 0 -c -o '+g_var.input_directory+'CG_INPUT.pdb', g_var.input_directory+'CG_INPUT.pdb'])
    if not os.path.exists(g_var.input_directory+'CG_INPUT.pdb'):
        sys.exit('\nFailed to process coarse-grained input file')      

#### gromacs parser
def gromacs(gro):
    possible_errors = ['File input/output error:','Error in user input:', 'did not converge to Fmax ', 
                        'but did not reach the requested Fmax ', 'Segmentation fault', 'Fatal error:', 'Cannot read from input', 'threads is only supported with thread-MPI']
    cmd,output = gro[0], gro[1]
    error = False
    if os.path.exists(output):
        pass
    else:
    #### if the flag gromacs is used every gromacs command will be printed to the terminal 
        if g_var.args.v >= 3:
            print('\nrunning gromacs: \n '+cmd+'\n')
        output = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        err, out = output.communicate()
        out=out.decode("utf-8")
        if not hasattr(g_var, 'gmx_version'):
            check_gromacs_version(out, err)
    #### all gromacs outputs will be saved into gromacs_outputs within the folder it is run
        with open('gromacs_outputs', 'a') as checks:
            checks.write(out)
    #### standard catch for failed gromacs commands
            for err in possible_errors:
                if err in out:
                    print('\n\nThere was an error with the following command:\n'+cmd+'\nThe exact error can be read in: '+os.getcwd()+'/gromacs_outputs\n')
                    if 'residue naming needs to be fixed' in out and 'PROTEIN_aligned' in out:
                        print('\n\n###  The supplied protein structure contains incorrectly named or missing atoms  ###\n\n')
                    error = True
#                if err == 'threads is only supported with thread-MPI':
#                    g_var.ntmpi = False
            if 'number of atoms in the topology (' in out:
                print('\n'+out+'\n\n')
                print('{0:^90}\n\n{1:^90}\n'.format('***NOTE***','If it is only out by multiples of two, check cysteine distances and increase -cys cutoff'))
                print('{0:^90}\n\n'.format('A lot of Martini v2-2 disulphide bonds can be up to 10 A (current search cutoff is '+str(g_var.args.cys)+' A)')) 
                error = True
    if len(gro) == 4: 
        gro[3].put(gro[2])
        return gro[2], error 
    return 0, error

def check_gromacs_version(output, err):
    err = err.decode("utf-8")
    for line in err.split('\n') :
        if line.startswith('GROMACS') and 'version' in line:
            if '-' in line.split()[-1].split('.')[0]:
                version = line.split()[-1].split('.')[0].split('-')[0]
                if version > 6:
                    g_var.gmx_version = True
            elif float(line.split()[-1].split('.')[0]) > 6:
                g_var.gmx_version = True
            else:
                g_var.gmx_version = False
                print('GROMACS version < 2016 detected therefore distance restraints will not be applied')


def make_min(residue):#, fragments):
#### makes minimisation folder
    gen.mkdir_directory('MIN')
#### makes em.mdp file for each residue
    if not os.path.exists('em_'+residue+'.mdp'):
        with open('em_'+residue+'.mdp','w') as em:
            em.write('define = \n integrator = steep\nnsteps = 20000\nemtol = 750\nemstep = 0.001\ncutoff-scheme = Verlet\n')

def pdb2gmx_minimise(chain,pdb2gmx_selections,res_type, q):
    os.chdir(g_var.working_dir+'/'+res_type)
    if not os.path.exists(res_type+'_de_novo_'+str(chain)+'_gmx.pdb'):
        pdb2gmx_chain(chain, 'de_novo_', res_type, ' << EOF \n1\n'+str(pdb2gmx_selections[chain][0])+'\n'+str(pdb2gmx_selections[chain][1]))
    if not os.path.exists(res_type+'_de_novo_'+str(chain)+'_gmx_checked.pdb'):
        at_mod.check_overlap_chain(chain, 'de_novo_', res_type)
    if g_var.user_at_input and not os.path.exists(res_type+'_aligned_'+str(chain)+'_gmx_checked.pdb') and res_type == 'PROTEIN':
        pdb2gmx_selections[chain] = histidine_protonation(chain, 'de_novo_', pdb2gmx_selections[chain])
        pdb2gmx_chain(chain, 'aligned_', res_type, pdb2gmx_selections[chain])
        at_mod.check_overlap_chain(chain, 'aligned_', res_type)
    minimise_protein_chain(chain, 'de_novo_', res_type)
    if g_var.user_at_input and res_type == 'PROTEIN':
        minimise_protein_chain(chain, 'aligned_', res_type)
    q.put(chain)
    return chain


def histidine_protonation(chain, input, chain_ter):
#### reads protonation state of histidine from itp file
    histidines=[]
    with open('PROTEIN_'+input+str(chain)+'.top', 'r') as top_input:
        for line in top_input.readlines():
            if line.startswith('; residue'):
                if line.split()[5] in ['HSD','HID', 'HISD']:
                    histidines.append(0)
                elif line.split()[5] in ['HSE', 'HIE', 'HISE']:
                    histidines.append(1)
                elif line.split()[5] in ['HSP','HIS1', 'HISP']:
                    histidines.append(2)
    pdb2gmx_selections='-his << EOF \n1'
    for his in histidines:
        pdb2gmx_selections+='\n'+str(his)
    pdb2gmx_selections+='\n'+str(chain_ter[0])+'\n'+str(chain_ter[1])
    return pdb2gmx_selections


### interactive terminal residue selection
def ask_ter_question(residue, options, chain):
    print('\n please select species for '+residue+' residue in chain '+str(chain))
    print('\nPlease select a option from below:\n')
    print('{0:^20}{1:^30}'.format('Selection','termini'))
    print('{0:^20}{1:^30}'.format('---------','----------'))
    sel=[]
    for selection, ter in enumerate(options):
        sel.append(ter)
        print('{0:^20}{1:^30}'.format(selection,ter))
    while True:
        try:
            number = int(input('\nplease select a option: '))
            if number < len(options):
                return number
        except KeyboardInterrupt:
            sys.exit('\nInterrupted')
        except BaseException:
            print("Oops!  That was a invalid choice")

def ask_terminal(sys_info, residue_type):
    system_ter=[]
    for chain in range(g_var.system[residue_type]):
        chain_ter=[]
        ter_name=['n','c']
        for ter_val,  ter_residue in enumerate(sys_info[chain]):
            if g_var.args.ter:
                chain_ter.append(ask_ter_question(g_var.res_top[ter_residue]['RESIDUE'][0], g_var.termini_selections[ter_name[ter_val]][g_var.res_top[ter_residue]['RESIDUE'][0]], chain))
            elif ter_residue in terminal_PTMs and ter_val == 0:
                # PTM with no modification
                chain_ter.extend(['6', '0'])
            elif len(chain_ter)!=2:
                chain_ter.append('0')
        system_ter.append(chain_ter)
    return system_ter

def run_parallel_pdb2gmx_min(res_type, sys_info):                        
    with mp.Pool(g_var.args.ncpus) as pool:
        m = mp.Manager()
        q = m.Queue()
        os.chdir(g_var.working_dir+res_type)
        make_min(res_type)
        gen.folder_copy_and_check(g_var.forcefield_location+g_var.forcefield, g_var.working_dir+res_type+'/'+g_var.forcefield+'/.')
        gen.file_copy_and_check(g_var.forcefield_location+g_var.forcefield+'/residuetypes.dat', g_var.working_dir+res_type+'/residuetypes.dat')
        pdb2gmx_selections=ask_terminal(sys_info, res_type)
        pool_process = pool.starmap_async(pdb2gmx_minimise, [(chain, pdb2gmx_selections,res_type, q) for chain in range(0, g_var.system[res_type])])
        while not pool_process.ready(): 
            report_complete('pdb2gmx/minimisation', q.qsize(), g_var.system[res_type])
    for chain in range(0, g_var.system[res_type]):
        if not os.path.exists(res_type+'_de_novo_'+str(chain)+'_gmx.pdb') or not os.path.exists(g_var.working_dir+res_type+'/MIN/'+res_type+'_de_novo_'+str(chain)+'.pdb'):
            print('For some reason parallisation of pdb2gmx failed on chain '+str(chain)+', now rerunning in serial.')
            pdb2gmx_minimise(chain, pdb2gmx_selections,res_type, q)
    print('{:<130}'.format(''), end='\r')
    print('\npdb2gmx/minimisation completed on residue type: '+res_type+'\n')

def pdb2gmx_chain(chain, input,res_type, pdb2gmx_selections):
#### pdb2gmx on on protein chain, creates the topologies
    ter = ' -ter '
    gromacs([g_var.args.gmx+' pdb2gmx -f '+res_type+'_'+input+str(chain)+'.pdb -o '+res_type+'_'+input+str(chain)+'_gmx.pdb -water none \
    -p '+res_type+'_'+input+str(chain)+'.top  -i '+res_type+'_'+str(chain)+'_posre.itp '+g_var.vs+ter+pdb2gmx_selections+'\nEOF', ''+res_type+'_'+input+str(chain)+'_gmx.pdb']) #### single chains
#### converts the topology file and processes it into a itp file
    convert_topology(res_type+'_'+input, chain, res_type)
#### writes topology overview for each chain 
    write_topol(res_type+'_'+input, 1, str(chain))
#### writes restraints file for each chain
    if res_type == 'PROTEIN':
        write_posres(chain)

def minimise_protein_chain(chain, input, res_type):
    #### grompps each protein chain
    gromacs([g_var.args.gmx+' grompp '+
                '-f em_'+res_type+'.mdp '+
                '-p topol_'+res_type+'_'+input+str(chain)+'.top '+
                '-c '+res_type+'_'+input+str(chain)+'_gmx_checked.pdb '+
                '-o MIN/'+res_type+'_'+input+str(chain)+' '+
                '-maxwarn 1 ', 'MIN/'+res_type+'_'+input+str(chain)+'.tpr'])
#### minimises chain
    os.chdir('MIN')
    # removed flag -nt 1
    gromacs([g_var.args.gmx+' mdrun -ntmpi 1 -v -deffnm '+res_type+'_'+input+str(chain)+' -c '+res_type+'_'+input+str(chain)+'.pdb', res_type+'_'+input+str(chain)+'.pdb'])
    os.chdir('..')  


def posres_header(file_write):
        posres_output = open(file_write, 'w')
        posres_output.write('[ position_restraints ]\n; atom  type      fx      fy      fz\n')
        return posres_output

def write_posres(chain):
#### if not posres file exist create one
    very_low_posres = posres_header(g_var.working_dir+'PROTEIN/PROTEIN_'+str(chain)+'_very_low_posre.itp')
    low_posres = posres_header(g_var.working_dir+'PROTEIN/PROTEIN_'+str(chain)+'_low_posre.itp')
    mid_posres = posres_header(g_var.working_dir+'PROTEIN/PROTEIN_'+str(chain)+'_mid_posre.itp')
    high_posres = posres_header(g_var.working_dir+'PROTEIN/PROTEIN_'+str(chain)+'_high_posre.itp')
    very_high_posres = posres_header(g_var.working_dir+'PROTEIN/PROTEIN_'+str(chain)+'_very_high_posre.itp')
    ultra_posres = posres_header(g_var.working_dir+'PROTEIN/PROTEIN_'+str(chain)+'_ultra_posre.itp')
    ca_posres = posres_header(g_var.working_dir+'PROTEIN/PROTEIN_'+str(chain)+'_ca_posre.itp')
    #### read in each chain from after pdb2gmx 
    with open(g_var.working_dir+'PROTEIN/PROTEIN_de_novo_'+str(chain)+'_gmx.pdb', 'r') as pdb_input:
        at_counter=0
        for line in pdb_input.readlines():
            if line.startswith('ATOM'):
                line_sep = gen.pdbatom(line)
                at_counter+=1
            #### if atom is in the restraint list for that residue add to position restraint file
                if line_sep['atom_name'] == 'CA':
                    ca_posres.write(str(at_counter)+'     1  1000  1000  1000\n')
                if not gen.is_hydrogen(line_sep['atom_name']):
                    very_low_posres.write(str(at_counter)+'     1  200  200  200\n')
                    low_posres.write(str(at_counter)+'     1  750  750  750\n')
                    mid_posres.write(str(at_counter)+'     1  1500  1500  1500\n')
                    high_posres.write(str(at_counter)+'     1  3000  3000  3000\n')
                    very_high_posres.write(str(at_counter)+'     1  6000  6000  6000\n')
                    ultra_posres.write(str(at_counter)+'     1  10000  10000  10000\n')

def convert_topology(topol, protein_number, res_type):
#### reads in topology 
    if os.path.exists(topol+str(protein_number)+'.top'):
        read=False
        mol_type=False
        if not os.path.exists(topol+str(protein_number)+'.itp'):
            print('no topology', protein_number, res_type)
            with open(topol+str(protein_number)+'.itp', 'w') as itp_write:

                for line in open(topol+str(protein_number)+'.top', 'r').readlines():
                    # print(line)
                #### copies between moleculetype and position restraint section
                    if len(line.split()) > 1: 
                        if read == False and line.split()[1] == 'moleculetype':
                            read = True
                        if line.split()[0]== '[' and line.split()[1] == 'moleculetype' and read:
                            mol_type = True
                        elif line.split()[0]== '[' and line.split()[1] == 'atoms' and read:
                            mol_type = False
                        elif mol_type:
                            if not line.startswith(';'):
                                line = '{0}       {1:20}'.format(res_type.lower()+'_'+str(protein_number), line.split()[1])
                        if read == True and line.split()[1] == 'POSRES':
                            read = False
                #### writes to itp file copied section          
                    if read:
                        # print(line)
                        itp_write.write(line)
                if res_type in ['PROTEIN']:
                #### adds position restraint section to end of itp file         
                    itp_write.write('#ifdef POSRES\n#include \"'+res_type+'_'+str(protein_number)+'_posre.itp\"\n#endif\n') 
                    itp_write.write('#ifdef POSRESCA\n#include \"'+res_type+'_'+str(protein_number)+'_ca_posre.itp\"\n#endif\n') 
                    itp_write.write('#ifdef VERY_LOWPOSRES\n#include \"'+res_type+'_'+str(protein_number)+'_very_low_posre.itp\"\n#endif\n')
                    itp_write.write('#ifdef LOWPOSRES\n#include \"'+res_type+'_'+str(protein_number)+'_low_posre.itp\"\n#endif\n')
                    itp_write.write('#ifdef MIDPOSRES\n#include \"'+res_type+'_'+str(protein_number)+'_mid_posre.itp\"\n#endif\n')
                    itp_write.write('#ifdef HIGHPOSRES\n#include \"'+res_type+'_'+str(protein_number)+'_high_posre.itp\"\n#endif\n')
                    itp_write.write('#ifdef VERY_HIGHPOSRES\n#include \"'+res_type+'_'+str(protein_number)+'_very_high_posre.itp\"\n#endif\n')
                    itp_write.write('#ifdef ULTRAPOSRES\n#include \"'+res_type+'_'+str(protein_number)+'_ultra_posre.itp\"\n#endif\n')
    else:
        sys.exit('cannot find : '+topol+'_'+str(protein_number)+'.top')

def write_topol(residue_type, residue_number, chain):
#### open topology file
    found=False
    with open('topol_'+residue_type+chain+'.top', 'w') as topol_write:
    #### add standard headers may need to be changed dependant on forcefield
        topol_write.write('; Include force field parameters\n#include \"../FINAL/'+g_var.forcefield+'/forcefield.itp\"\n')
    #### add location of residue topology file absolute file locations
        if residue_type in g_var.sol_residues+g_var.ion_residues+g_var.np_residues:
            for directory in g_var.sol_directories+g_var.ion_directories+g_var.np_directories:
                if os.path.exists(directory[0]+residue_type+'/'+gen.swap_to_solvent(residue_type)+'.itp'):
                    topol_write.write('#include \"'+directory[0]+residue_type+'/'+gen.swap_to_solvent(residue_type)+'.itp\"')
                    found=True
            if not found:
                sys.exit('cannot find itp : '+residue_type+'/'+gen.swap_to_solvent(residue_type)+'.itp')
        elif os.path.exists(g_var.working_dir+'/'+residue_type.split('_')[0]+'/'+residue_type+chain+'.itp'):
            topol_write.write('#include \"'+residue_type+chain+'.itp\"\n')
        else:
            sys.exit('cannot find itp : '+residue_type+'/'+residue_type+chain)
    #### topology section headers
        topol_write.write('\n\n[ system ]\n; Name\nSomething clever....\n\n[ molecules ]\n; Compound        #mols\n')
    #### individual number of residues
        if residue_type.split('_')[0] in ['PROTEIN', 'OTHER']:
             residue_type=residue_type.split('_')[0]+'_'
        residue_type = gen.swap_to_solvent(residue_type)
        topol_write.write(residue_type+chain+'    '+str(residue_number))


#################################################################   Non protein

def report_complete(func, size, resid):
    if np.round((size/resid)*100,2).is_integer():
        print('Running '+func+' on '+str(resid)+' molecules. Percentage complete: ',np.round((size/resid)*100,2),'%', end='\r')
    time.sleep(0.1)

def minimise_merged(residue_type, input_file):
#### write topology for merged system    
    os.chdir(g_var.working_dir+residue_type)
    make_min(residue_type)
    write_topol(residue_type, g_var.system[residue_type], '')
#### grompp with merged system
    gromacs([g_var.args.gmx+' grompp '+
            '-po md_out-'+residue_type+' '+
            '-f em_'+residue_type+'.mdp '+
            '-p topol_'+residue_type+'.top '+
            '-c '+input_file+' '+
            '-o '+g_var.working_dir+residue_type+'/MIN/'+residue_type+'_merged_min -maxwarn 1', g_var.working_dir+residue_type+'/MIN/'+residue_type+'_merged_min.tpr'])
#### change to min directory and minimise
    os.chdir('MIN') 
    # removed flag -nt '+str(g_var.args.ncpus)+'
    complete, success = gromacs([g_var.args.gmx+' mdrun -ntmpi 1 -v -pin on -deffnm '+residue_type+'_merged_min -c ../'+residue_type+'_merged.pdb', '../'+residue_type+'_merged.pdb'])
    os.chdir(g_var.working_dir)
    return success

################################################################ Gromacs for merged system

def check_atom_type(line, a_line, atomtypes_itp_lines):
    line_sep=line.split() 
    name = int(np.where(line_sep[0]==a_line[:,0])[0])  
    bond = int(np.where(line_sep[1]==a_line[:,1])[0]) 
    if name == bond: 
        if float(line_sep[5]) != float(a_line[name][5]) or float(line_sep[6]) != float(a_line[name][6]): 
            sys.exit('\nThere are duplicate atomtypes in your molecules: \n'+line) 

def strip_atomtypes(itp_file): 
    with open(itp_file, 'r') as itp_input: 
        itp_lines = itp_input.read().splitlines() 
    if '[ atomtypes ]' in itp_lines: 
        a_lines_sep = []
        if not os.path.exists('extra_atomtypes.itp'): 
            atomtypes_output = open('extra_atomtypes.itp', 'w') 
            atomtypes_output.write('[ atomtypes ]\n') 
        else: 
            atomtypes_output = open('extra_atomtypes.itp', 'a') 
            with open('extra_atomtypes.itp', 'r') as atomtypes_itp_r: 
                atomtypes_itp_lines = atomtypes_itp_r.read().splitlines() 
            for a_line in atomtypes_itp_lines[1:]: 
                if not a_line.startswith(';'):  
                    a_lines_sep.append(a_line.split()) 
            a_lines_sep = np.array(a_lines_sep) 
        atom = itp_lines.index('[ atomtypes ]')+1 
        mol = itp_lines.index('[ moleculetype ]') 
        for line in itp_lines[atom:mol]: 
            if not line.startswith(';'): 
                line_sep = line.split() 
                if len(line_sep) > 4: 
                    if len(a_lines_sep) > 2:
                        if line_sep[0] not in a_lines_sep[:,0] and line_sep[1] not in a_lines_sep[:,1]: 
                            atomtypes_output.write(line+'\n') 
                        else: 
                            check_atom_type(line, a_lines_sep, atomtypes_itp_lines) 
                    else:
                        atomtypes_output.write(line+'\n')
        with open(itp_file, 'w') as itp_output: 
            for line in itp_lines[mol:]: 
                itp_output.write(line+'\n')



def write_merged_topol():
    os.chdir(g_var.working_dir+'MERGED')
    # if not os.path.exists('topol_final.top'):
    with open('topol_final.top', 'w') as topol_write:
        topologies_to_include=[]
    #### runs through residue types and copies to MERGED directory and simplifies the names
        for residue_type in g_var.system:
            #### copies 1st itp file it comes across 
            for directory in g_var.np_directories+g_var.sol_directories+g_var.ion_directories:
                residue_type_name = gen.swap_to_solvent(residue_type)
                if os.path.exists(directory[0]+residue_type+'/'+residue_type_name+'.itp'):  
                    if not any(residue_type_name+'.itp' in s for s in topologies_to_include):
                        topologies_to_include.append('#include \"'+residue_type_name+'.itp\"\n')
                        gen.file_copy_and_check(directory[0]+residue_type+'/'+residue_type_name+'.itp', residue_type_name+'.itp')
                        gen.file_copy_and_check(directory[0]+residue_type+'/'+residue_type_name+'_posre.itp', residue_type_name+'_posre.itp')
                        strip_atomtypes(residue_type_name+'.itp')
                    break
        #### copies across protein itp files and simplifies the names 
            if residue_type in ['PROTEIN', 'OTHER']:
                for unit in range(g_var.system[residue_type]): 
                    topologies_to_include.append('#include \"'+residue_type+'_'+str(unit)+'.itp\"\n')
                    gen.file_copy_and_check(g_var.working_dir+residue_type+'/'+residue_type+'_de_novo_'+str(unit)+'.itp', residue_type+'_'+str(unit)+'.itp')
                    if residue_type in ['PROTEIN']:
                        for posres_type in ['_very_low_posre.itp','_low_posre.itp','_mid_posre.itp','_high_posre.itp','_very_high_posre.itp','_ultra_posre.itp','_ca_posre.itp','_posre.itp']:
                            gen.file_copy_and_check(g_var.working_dir+'PROTEIN/PROTEIN_'+str(unit)+posres_type, 'PROTEIN_'+str(unit)+posres_type)
                        gen.file_copy_and_check(g_var.working_dir+'PROTEIN/PROTEIN_disres.itp', 'PROTEIN_disres.itp')  
        if os.path.exists('extra_atomtypes.itp'):
            topol_write.write('; Include force field parameters\n#include \"../FINAL/'+g_var.forcefield+'/forcefield.itp\"\n')
            topol_write.write('#include \"extra_atomtypes.itp\"\n')
        else:
            topol_write.write('; Include force field parameters\n#include \"../FINAL/'+g_var.forcefield+'/forcefield.itp\"\n')
        for line in topologies_to_include:
            topol_write.write(line)

        topol_write.write('[ system ]\n; Name\nSomething clever....\n\n[ molecules ]\n; Compound        #mols\n')
    #### adds number of residues to the topology
        for residue_type in g_var.system:
            if residue_type not in  ['PROTEIN', 'OTHER']:
                topol_write.write(gen.swap_to_solvent(residue_type)+'    '+str(g_var.system[residue_type])+'\n')   
        #### adds monomers separately
            else:
                for unit in range(g_var.system[residue_type]):
                    topol_write.write(residue_type+'_'+str(unit)+'    1\n')    
        topol_write.write('\n#ifdef DISRES\n#include \"PROTEIN_disres.itp\"\n#endif')


def minimise_merged_pdbs(protein):
    print('\nMinimising merged atomistic files : '+protein[1:])
    os.chdir(g_var.working_dir+'MERGED')
#### grompps final merged systems
    gromacs([g_var.args.gmx+' grompp '+
            '-po md_out-merged_cg2at '+
            '-f em_merged_cg2at.mdp '+
            '-p topol_final.top '+
            '-r merged_cg2at'+protein+'.pdb '+
            '-c merged_cg2at'+protein+'.pdb '+
            '-o MIN/merged_cg2at'+protein+'_minimised '+
            '-maxwarn 1', 'MIN/merged_cg2at'+protein+'_minimised.tpr'])
    os.chdir('MIN')
#### runs minimises final systems
# removed flag -nt '+str(g_var.args.ncpus)+'
    gromacs([g_var.args.gmx+' mdrun -ntmpi 1 -v -pin on -deffnm merged_cg2at'+protein+'_minimised -c merged_cg2at'+protein+'_minimised.pdb', 'merged_cg2at'+protein+'_minimised.pdb'])

   
def write_steered_mdp(loc, posres, time, timestep):
    if not os.path.exists(loc):
        with open(loc, 'w') as steered_md:
            steered_md.write('define = '+posres+'\nintegrator = md\nnsteps = '+str(time)+'\ndt = '+str(timestep)+'\ncontinuation   = no\nconstraint_algorithm = lincs\n')
            steered_md.write('nstxout-compressed = 1000\nnstenergy = 1000\nconstraints = h-bonds\nnstlist = 25\nrlist = 1.2\nrcoulomb = 1.2\nrvdw = 1.2\ncoulombtype  = PME\n')
            steered_md.write('pme_order = 4\nfourierspacing = 0.135\ntcoupl = v-rescale\ntc-grps = system\ntau_t = 0.1\nref_t = 310\npcoupl = no\n')
            steered_md.write('pbc = xyz\nDispCorr = no\ngen_vel = no\nrefcoord_scaling = all\ncutoff-scheme = Verlet\n')
            steered_md.write('disre=simple\ndisre-weighting=equal\ndisre-fc=1000\ndisre-tau=1\nnstdisreout=1\n')   

def steer_to_aligned(protein_type, fc, input_file ):
    gen.mkdir_directory(g_var.merged_directory+'STEER')
    print('Applying '+fc.replace('_',' ')+' position restraints', end='\r')
    os.chdir(g_var.merged_directory)
    write_steered_mdp(g_var.merged_directory+fc+'_posres-nvt.mdp', '-D'+fc.upper()+'POSRES -DNP', 2000, 0.001)  
    gromacs([g_var.args.gmx+' grompp '+
            ' -po md_out-merged_cg2at_steer_'+fc+
            ' -f '+fc+'_posres-nvt.mdp '+
            ' -p topol_final.top '+
            ' -r merged_cg2at_'+protein_type+'.pdb '+
            ' -c '+input_file+'.pdb '+
            ' -o STEER/merged_cg2at_'+protein_type+'_steer_'+fc+' '+
            ' -maxwarn '+str(2), 'STEER/merged_cg2at_'+protein_type+'_steer_'+fc+'.tpr'])  
    os.chdir('STEER')
    # removed flag -nt '+str(g_var.args.ncpus)+'
    complete, equil = gromacs([g_var.args.gmx+' mdrun -ntmpi 1 -v -pin on -deffnm merged_cg2at_'+protein_type+'_steer_'+fc+
                                 ' -c merged_cg2at_'+protein_type+'_steer_'+fc+'.pdb -cpo merged_cg2at_'+protein_type+'_steer_'+fc+'.cpt'
                                 ,'merged_cg2at_'+protein_type+'_steer_'+fc+'.pdb'])
    print('{:<100}'.format(''), end='\r')
    return equil


def run_nvt(input_file):
    print('\nRunning NVT on de novo system', end='\r')
    os.chdir(g_var.merged_directory)   
    gen.mkdir_directory(g_var.merged_directory+'NVT')
    if g_var.user_at_input and g_var.args.disre and g_var.gmx_version:
        write_steered_mdp(g_var.merged_directory+'nvt.mdp', '-DDISRES -DPOSRESCA', 5000, 0.001)
    elif 'PROTEIN' in g_var.system:
        write_steered_mdp(g_var.merged_directory+'nvt.mdp', '-DPOSRESCA', 5000, 0.001)
    else:
        write_steered_mdp(g_var.merged_directory+'nvt.mdp', '', 5000, 0.001)
    gromacs([g_var.args.gmx+' grompp'+
            ' -po md_out-merged_cg2at_nvt'+
            ' -f nvt.mdp'+
            ' -p topol_final.top'+
            ' -r '+input_file+'.pdb '+
            ' -c '+input_file+'.pdb '+
            ' -o NVT/merged_cg2at_de_novo_nvt'+
            ' -maxwarn '+str(2), 'NVT/merged_cg2at_de_novo_nvt.tpr'])   
    os.chdir(g_var.merged_directory+'NVT')
    # removed flag -nt '+str(g_var.args.ncpus)+'
    gromacs([g_var.args.gmx+' mdrun -ntmpi 1 -v -pin on -deffnm merged_cg2at_de_novo_nvt'+
            ' -c merged_cg2at_de_novo_nvt.pdb -cpo merged_cg2at_de_novo_nvt.cpt'
            , 'merged_cg2at_de_novo_nvt.pdb'])  
    gen.file_copy_and_check('merged_cg2at_de_novo_nvt.pdb', g_var.final_dir+'final_cg2at_de_novo.pdb')    
    print('Completed NVT, please find final de_novo system: \n'+g_var.final_dir+'final_cg2at_de_novo.pdb')

def create_aligned():
    print('\nCreating aligned system') 
    at_mod.merge_system_pdbs('_aligned') ## create restraint positions for aligned system
    aligned_atoms, chain_count = read_in.read_in_atomistic(g_var.working_dir+'PROTEIN/PROTEIN_aligned_merged.pdb') ## reads in final pdb
    rmsd = at_mod_p.RMSD_measure_de_novo(aligned_atoms) ## gets rmsd of de novo
    for chain in rmsd:
        if rmsd[chain] > 3:
            print('Your aligned structure is quite far from the CG, therefore running gentle steering\n')
            print_rmsd(rmsd)
            steer = ['very_low', 'low', 'mid', 'high', 'very_high', 'ultra']
            break
        else:
            steer = ['low', 'high', 'ultra']

    final_file = run_steer(steer, g_var.merged_directory+'checked_ringed_lipid_de_novo')
    if final_file:
        gen.file_copy_and_check(final_file, g_var.final_dir+'final_cg2at_aligned.pdb') ## copy to final folder
    else:
        final_file = run_steer(['very_low', 'low', 'mid', 'high', 'very_high', 'ultra'], g_var.merged_directory+'checked_ringed_lipid_de_novo')
        print('Completed alignment, please find final aligned system: \n'+g_var.final_dir+'final_cg2at_aligned.pdb')
        gen.file_copy_and_check(final_file, g_var.final_dir+'final_cg2at_aligned.pdb') ## copy to final folder

def print_rmsd(rmsd):
    print('\n{0:^25}{1:^10}'.format('chain','RMSD ('+chr(197)+')'))
    print('{0:^25}{1:^10}'.format('-----','---------'))
    for key, rmsd_val in rmsd.items(): print('{0:^25}{1:^10}'.format(str(key), float(rmsd_val)))
    print()

def run_steer(steer, initial):
    for res_val, restraint in enumerate(steer):
        if not os.path.exists(g_var.merged_directory+'STEER/merged_cg2at_aligned_steer_'+restraint+'pdb'):
            if res_val == 0:
                equil = steer_to_aligned('aligned', restraint, initial)
            else:
                equil = steer_to_aligned('aligned', restraint, g_var.merged_directory+'STEER/merged_cg2at_aligned_steer_'+steer[res_val-1])
            if equil:
                print('Steering to aligned failed at: '+restraint)
                if len(steer) > 5:
                    print('Your aligned structure may be too far from the CG input')
                    print('The closest the script can get, is found in the FINAL directory')
                    return g_var.merged_directory+'STEER/merged_cg2at_aligned_steer_'+steer[res_val]+'.pdb'
                else:
                    print('Your aligned structure may be quite far from the CG input')
                    print('Restarting with gentler restraints')
                    rmtree(g_var.merged_directory+'STEER')
                    return False 
    return g_var.merged_directory+'STEER/merged_cg2at_aligned_steer_'+steer[-1]+'.pdb'
